#include <Windows.h>
#include <gl/GL.h>
#include <ReWindow/types/Window.h>
#include <ReWindow/Logger.h>


class ReWindow::RWindow::Platform {
public:
    HINSTANCE hInstance;
    HWND hwnd;
    HDC hdc;
};

using namespace ReWindow;

//bool local_fullscreen_mode = false;
bool local_focused = true;
//bool local_vsync = false;
//bool local_cursor_visible = true;
//bool local_closing = false;
void RWindow::PollEvents() {
    MSG msg;
    while (PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE))
        TranslateMessage(&msg),
        DispatchMessage(&msg);
    focused = local_focused;
}

LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    auto* window = reinterpret_cast<RWindow*>( GetWindowLongPtr(hwnd, GWLP_USERDATA) );
    switch (uMsg) {
            case WM_CLOSE: {
                window->processOnClose();
                DestroyWindow(hwnd);

            }

            case WM_DESTROY: {
                exit(0);
            }

            case WM_SIZE: {
                window->SetSizeWithoutEvent({ LOWORD(lParam), HIWORD(lParam) });
                auto eventData = WindowResizeRequestEvent();
                eventData.Size = { window->GetWidth(), window->GetHeight() };
                window->SetLastKnownWindowSize({ window->GetWidth(), window->GetHeight() });
                // TODO: Implement eWindow->processOnResize()
                window->OnResizeRequest(eventData);
                window->OnResizeRequestEvent(eventData);
                break;
            }

            case WM_SETFOCUS: {
                window->processFocusIn();
                local_focused = true;

                // TODO actually check if it's flashing.
                FLASHWINFO fi;
                fi.cbSize = sizeof(FLASHWINFO);
                fi.hwnd = hwnd;
                fi.dwFlags = FLASHW_STOP;
                fi.uCount = 0;
                fi.dwTimeout = 0;
                FlashWindowEx(&fi);
                break;
            }

            case WM_KILLFOCUS: {
                window->processFocusOut();
                local_focused = false;
            }

            case WM_SETCURSOR: {
                if (LOWORD(lParam) == HTCLIENT && window->GetCursorVisible() == false)
                    SetCursor(nullptr);
                break;
            }

            case WM_KEYDOWN: {
                auto key = GetKeyFromWindowsScancode((WindowsScancode) wParam);
                //Key repeat fix.
                if (!window->previousKeyboard.PressedKeys[key])
                    window->processKeyPress(key);
                break;
            }

            case WM_KEYUP: {
                auto key = GetKeyFromWindowsScancode((WindowsScancode) wParam);
                window->processKeyRelease(key);
                break;
            }

            //Mouse Buttons.
            case WM_MOUSEWHEEL: {
                int wheel_delta = GET_WHEEL_DELTA_WPARAM(wParam);
                // Clamp to normalized range [+1 or -1].
                // TODO: Test this and make sure sign is the same on both platforms for each direction.
                wheel_delta = (wheel_delta > 0) ? 1 : -1;
                // TODO: Determine sign of wheel_delta for each direction, (and on linux too), and document this.
                window->processMouseWheel(wheel_delta);
                break;
            }

            case WM_LBUTTONDOWN: {
                window->processMousePress(MouseButtons::Left);
                break;
            }

            case WM_LBUTTONUP: {
                window->processMouseRelease(MouseButtons::Left);
                break;
            }

            case WM_RBUTTONDOWN: {
                window->processMousePress(MouseButtons::Right);
                break;
            }

            case WM_RBUTTONUP: {
                window->processMouseRelease(MouseButtons::Right);
                break;
            }

            case WM_MBUTTONDOWN: {
                window->processMousePress(MouseButtons::Middle);
                break;
            }

            case WM_MBUTTONUP: {
                window->processMouseRelease(MouseButtons::Middle);
                break;
            }

            case WM_XBUTTONDOWN: {
                WORD button = GET_XBUTTON_WPARAM(wParam);
                if (button == XBUTTON1)
                    window->processMousePress(MouseButtons::Mouse4);

                if (button == XBUTTON2)
                    window->processMousePress(MouseButtons::Mouse5);
                break;
            }

            case WM_XBUTTONUP: {
                WORD button = GET_XBUTTON_WPARAM(wParam);
                if (button == XBUTTON1)
                    window->processMouseRelease(MouseButtons::Mouse4);

                if (button == XBUTTON2)
                    window->processMouseRelease(MouseButtons::Mouse5);
                break;
        }

        //This is the same as "Motion Notify" in the X Window System.
        case WM_MOUSEMOVE:
            break;
    }
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

bool fullscreenmode = false;
bool open = false;

void RWindow::Raise() { SetWindowPos(platform->hwnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE); }
void RWindow::Lower() { SetWindowPos(platform->hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE); }

//TODO somehow check if our window is entirely obscured by other windows.
bool RWindow::IsVisible() const {
    if (!IsWindow(platform->hwnd))
        return false;
    if (IsIconic(platform->hwnd))
        return false;
    return true;
}

void RWindow::Flash() {
    if (!focused) {
        FLASHWINFO fi;
        fi.cbSize = sizeof(FLASHWINFO);
        fi.hwnd = platform->hwnd;
        fi.dwFlags = FLASHW_ALL;
        fi.uCount = 0;
        fi.dwTimeout = 0;
        FlashWindowEx(&fi);
    }
}

void RWindow::DestroyOSWindowHandle() {
    DestroyWindow(platform->hwnd);
}

void RWindow::SetResizable(bool resizable) {
    if (!resizable) {
        RECT rect;
        GetWindowRect(platform->hwnd, &rect);
        LONG style = GetWindowLong(platform->hwnd, GWL_STYLE);
        style &= ~(WS_THICKFRAME | WS_MAXIMIZEBOX);
        SetWindowLong(platform->hwnd, GWL_STYLE, style);
        SetWindowPos(platform->hwnd, nullptr, rect.left, rect.top, rect.right - rect.left,
        rect.bottom - rect.top, SWP_FRAMECHANGED | SWP_NOZORDER | SWP_NOOWNERZORDER);
    }
}

void RWindow::SetSize(int newWidth, int newHeight) {
    if (!resizable) return;

    this->width = newWidth;
    this->height = newHeight;
    SetWindowPos(platform->hwnd, nullptr, 0, 0, width, height, SWP_NOMOVE | SWP_NOZORDER);
}

IPair RWindow::GetAccurateMouseCoordinates() const {
    POINT point;
    GetCursorPos(&point);
    ScreenToClient(platform->hwnd, &point);
    return { point.x, point.y };
}

void RWindow::SetCursorVisible(bool cursor_enable) {
    cursor_visible = cursor_enable;
}

bool RWindow::GetCursorVisible() {
    return cursor_visible;
}


IPair RWindow::GetSize() const {
    RECT rect;
    GetClientRect(platform->hwnd, &rect);
    return { (rect.right - rect.left), (rect.bottom - rect.top) };
}

void RWindow::SetPos(int x, int y) {
    SetWindowPos(platform->hwnd, nullptr, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

void RWindow::SetPos(const IPair& pos) {
    SetPos(pos.x, pos.y);
}

void RWindow::Fullscreen() {
    // Implement fullscreen
    SetWindowLong(platform->hwnd, GWL_STYLE, GetWindowLong(platform->hwnd, GWL_STYLE) & ~WS_OVERLAPPEDWINDOW);
    SetWindowPos(platform->hwnd, HWND_TOP, 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN), SWP_FRAMECHANGED | SWP_NOOWNERZORDER);
}

void RWindow::RestoreFromFullscreen() {
    // Implement restore from fullscreen
    SetWindowLong(platform->hwnd, GWL_STYLE, GetWindowLong(platform->hwnd, GWL_STYLE) | WS_OVERLAPPEDWINDOW);
    SetWindowPos(platform->hwnd, nullptr, 0, 0, width, height, SWP_FRAMECHANGED | SWP_NOZORDER | SWP_NOOWNERZORDER);
}

void RWindow::SetCursorStyle(CursorStyle style) const {}

RWindow::RWindow() {
    extant = this;
    platform = new Platform();
}

RWindow::RWindow(const std::string& wTitle, int wWidth, int wHeight,
bool wFullscreen, bool wResizable, bool wVsync) :
title(wTitle), width(wWidth), height(wHeight), fullscreen_mode(wFullscreen), resizable(wResizable),
vsync(wVsync) {
    extant = this;
    platform = new Platform();
}

void RWindow::UniqueFunctionNameForMessageBoxBecauseMicrosoftUsesMacros(const std::string& title, const std::string& message_box_text) {
    MessageBox(nullptr, message_box_text.c_str(), title.c_str(), MB_OK);
}

HGLRC glContext;
HMODULE gl_lib = nullptr;
bool constructor_success = true;

namespace OpenGL {
    typedef const GLubyte* (APIENTRY *GLGETSTRINGPROC)(GLenum name);
    GLGETSTRINGPROC glGetString = nullptr;

    typedef BOOL (WINAPI *WGLMAKECURRENTPROC)(HDC hdc, HGLRC hglrc);
    WGLMAKECURRENTPROC wglMakeCurrent = nullptr;

    typedef FARPROC (WINAPI *WGLGETPROCADDRPROC)(LPCSTR lpszProc);
    WGLGETPROCADDRPROC wglGetProcAddress = nullptr;


    typedef HGLRC (WINAPI *WGLCREATECONTEXTPROC)(HDC hdc);
    WGLCREATECONTEXTPROC wglCreateContext = nullptr;

}

bool OpenGLWindow::Open() {
    if (!constructor_success)
        return false;

  platform->hInstance = GetModuleHandle(nullptr);

  WNDCLASS wc = { };
  wc.lpfnWndProc = WindowProc;
  wc.hInstance = platform->hInstance;
  wc.lpszClassName = "RWindowClass";
  RegisterClass(&wc);
  platform->hwnd = CreateWindowEx(
      0,
      "RWindowClass",
      title.c_str(),
      WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, width, height,
      nullptr,
      nullptr,
      platform->hInstance,
      nullptr
  );
    SetWindowLongPtr(platform->hwnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this));
  PIXELFORMATDESCRIPTOR pfd = {
      sizeof(PIXELFORMATDESCRIPTOR),
      1,
      PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
      PFD_TYPE_RGBA,
      24,
      0, 0, 0, 0, 0, 0,
      0,
      0,
      0,
      0, 0, 0, 0,
      24,
      8,
      0,
      PFD_MAIN_PLANE,
      0,
      0, 0, 0
  };

  platform->hdc = GetDC(platform->hwnd);
  int pixelFormat = ChoosePixelFormat(platform->hdc, &pfd);
  SetPixelFormat(platform->hdc, pixelFormat, &pfd);
  glContext = OpenGL::wglCreateContext(platform->hdc);
  OpenGL::wglMakeCurrent(platform->hdc, glContext);
  ShowWindow(platform->hwnd, SW_SHOW);
  open = true;
    return true;
}

std::string OpenGLWindow::GetGraphicsDriverVendor() {
    return std::string(reinterpret_cast<const char*>(OpenGL::glGetString(GL_VENDOR)));
}

void OpenGLWindow::SwapBuffers() {
    ::SwapBuffers(platform->hdc);
}

void OpenGLWindow::SetVsyncEnabled(bool b) {
    typedef BOOL(WINAPI* PFNWGLSWAPINTERVALEXTPROC)(int interval);
    auto wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC) OpenGL::wglGetProcAddress("wglSwapIntervalEXT");

    if (wglSwapIntervalEXT)
        wglSwapIntervalEXT(b ? 1 : 0);
}

bool OpenGLWindow::SoftwareRendered() {
    std::string renderer(reinterpret_cast<const char*>(OpenGL::glGetString(GL_RENDERER)));
    if (renderer.find("llvmpipe"))
        return true;
    if (renderer.find("softpipe"))
        return true;
    if (renderer.find("GDI Generic"))
        return true;
    if (GetGraphicsDriverVendor().find("Mesa"))
        return true;
    return false;
}

OpenGLWindow::OpenGLWindow(const std::string& title, int width, int height, uint8_t gl_major, uint8_t gl_minor)
: gl_major(gl_major), gl_minor(gl_minor), RWindow(title, width, height) {
    gl_lib = LoadLibrary("opengl32.dll");

    if (!gl_lib)
        Logger::Error("opengl32.dll couldn't be found. If you have it, you can try putting it in the same directory as the executable."),
        constructor_success = false;

    OpenGL::glGetString = (OpenGL::GLGETSTRINGPROC) GetProcAddress(gl_lib, "glGetString");
    OpenGL::wglMakeCurrent = (OpenGL::WGLMAKECURRENTPROC) GetProcAddress(gl_lib, "wglMakeCurrent");
    OpenGL::wglGetProcAddress = (OpenGL::WGLGETPROCADDRPROC) GetProcAddress(gl_lib, "wglGetProcAddress");
    OpenGL::wglCreateContext = (OpenGL::WGLCREATECONTEXTPROC) GetProcAddress(gl_lib, "wglCreateContext");
}
